import os
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib as mpl

'''
    This file contains a set of functions that are useful in analyzing visual behavior data
'''


def save_figure(fig, figsize, save_dir, folder, filename, formats=['.png']):
    '''
        Function for saving a figure
    
        INPUTS:
        fig: a figure object
        figsize: tuple of desired figure size
        save_dir: string, the directory to save the figure
        folder: string, the sub-folder to save the figure in. if the folder does not exist, it will be created
        filename: string, the desired name of the saved figure
        formats: a list of file formats as strings to save the figure as, ex: ['.png','.pdf']
    '''
    fig_dir = os.path.join(save_dir, folder)
    if not os.path.exists(fig_dir):
        os.mkdir(fig_dir)
    mpl.rcParams['pdf.fonttype'] = 42
    fig.set_size_inches(figsize)
    for f in formats:
        fig.savefig(os.path.join(fig_dir, fig_title + f), transparent=True, orientation='landscape')


def get_dff_matrix(session):
    '''
        Returns the dff_trace of a session as a numpy matrix

        INPUTS:
        session: a behaviorOphysSession object
        
        OUTPUTS:
        dff: a matrix of cells x dff_trace for the entire session
    '''
    dff = np.stack(session.dff_traces.dff, axis=0)
    return dff


def get_mean_df(response_df, conditions=['cell_specimen_id', 'image_name']):
    '''
        Computes an analysis on a selection of responses (either flashes or trials). Computes mean_response, sem_response, the pref_stim, fraction_active_responses.

        INPUTS
        response_df: the dataframe to group
        conditions: the conditions to group by, the first entry should be 'cell_specimen_id', the second could be 'image_name' or 'change_image_name'

        OUTPUTS:
        mdf: a dataframe with the following columns:
            mean_response: the average mean_response for each condition
            sem_response: the sem of the mean_response
            mean_trace: the average dff trace for each condition
            sem_trace: the sem of the mean_trace
            mean_responses: the list of mean_responses for each element of each group
            pref_stim: if conditions includes image_name or change_image_name, sets a boolean column for whether that was the cell's preferred stimulus
            fraction_significant_responses: the fraction of individual image presentations or trials that were significant (p_value > 0.05)
    '''

    # Group by conditions
    rdf = response_df.copy()
    mdf = rdf.groupby(conditions).apply(get_mean_sem_trace)
    mdf = mdf[['mean_response', 'sem_response', 'mean_trace', 'sem_trace', 'mean_responses']]
    mdf = mdf.reset_index()

    # Add preferred stimulus if we can
    if ('image_name' in conditions) or ('change_image_name' in conditions):
        mdf = annotate_mean_df_with_pref_stim(mdf)

    # What fraction of individual responses were significant?
    fraction_significant_responses = rdf.groupby(conditions).apply(get_fraction_significant_responses)
    fraction_significant_responses = fraction_significant_responses.reset_index()
    mdf['fraction_significant_responses'] = fraction_significant_responses.fraction_significant_responses

    if 'index' in mdf.keys():
        mdf = mdf.drop(columns=['index'])
    return mdf


def get_mean_sem_trace(group):
    '''
        Computes the average and sem of the mean_response column

        INPUTS:
        group: a pandas groupby object
        
        OUTPUT:
        a pandas series with the mean_response, sem_response, mean_trace, sem_trace, and mean_responses computed for the group. 
    '''
    mean_response = np.mean(group['mean_response'])
    mean_responses = group['mean_response'].values
    sem_response = np.std(group['mean_response'].values) / np.sqrt(len(group['mean_response'].values))
    mean_trace = np.mean(group['dff_trace'])
    sem_trace = np.std(group['dff_trace'].values) / np.sqrt(len(group['dff_trace'].values))
    return pd.Series({'mean_response': mean_response, 'sem_response': sem_response,
                      'mean_trace': mean_trace, 'sem_trace': sem_trace,
                      'mean_responses': mean_responses})


def annotate_mean_df_with_pref_stim(mean_df):
    '''
        Computes the preferred stimulus for each cell/trial or cell/flash combination. Preferred image is computed by seeing which image evoked the largest average mean_response across all images. 

        INPUTS:
        mean_df: the mean_df to be annotated

        OUTPUTS:
        mean_df with a new column appended 'pref_stim' which is a boolean TRUE/FALSE for whether that image was that cell's preferred image.
       
        ASSERTS:
        Each cell has one unique preferred stimulus 
    '''

    # Are we dealing with flash_response or trial_response
    if 'image_name' in mean_df.keys():
        image_name = 'image_name'
    else:
        image_name = 'change_image_name'

    # set up dataframe
    mdf = mean_df.reset_index()
    mdf['pref_stim'] = False

    # Iterate through cells in df       
    for cell in mdf['cell_specimen_id'].unique():
        mc = mdf[(mdf['cell_specimen_id'] == cell)]
        mc = mc[mc[image_name] != 'omitted']
        temp = mc[(mc.mean_response == np.max(mc.mean_response.values))][image_name].values
        if len(temp) > 0:  # need this test if the mean_response was nan
            pref_image = temp[0]
            # PROBLEM, this is slow, and sets on slice, better to use mdf.at[test, 'pref_stim']
            row = mdf[(mdf['cell_specimen_id'] == cell) & (mdf[image_name] == pref_image)].index
            mdf.loc[row, 'pref_stim'] = True

    # Test to ensure preferred stimulus is unique for each cell
    for cell in mdf.reset_index()['cell_specimen_id'].unique():
        if image_name == 'image_name':
            assert len(
                mdf.reset_index().set_index('cell_specimen_id').loc[cell].query('pref_stim').image_name.unique()) == 1
        else:
            assert len(mdf.reset_index().set_index('cell_specimen_id').loc[cell].query(
                'pref_stim').change_image_name.unique()) == 1
    return mdf


def get_fraction_significant_responses(group, threshold=0.05):
    '''
        Calculates the fraction of trials or flashes that have a p_value below threshold
        Note that this function does not handle multiple comparisons
    
        INPUT:
        group: a pandas groupby object
        threshold: the p_value threshold for significance for an individual response

        OUTPUT:
        a pandas series with column 'fraction_significant_responses'
    '''
    fraction_significant_responses = len(group[group.p_value < threshold]) / float(len(group))
    return pd.Series({'fraction_significant_responses': fraction_significant_responses})


def get_xticks_xticklabels(trace, ophys_frame_rate=31., interval_sec=1, window=[-4, 8]):
    """
    Function that accepts a timeseries, evaluates the number of points in the trace,
    and converts from acquisition frames to timestamps relative to a given window of time covered by the trace.

    :param trace: a single trace where length = the number of timepoints
    :param ophys_frame_rate: ophys frame rate if plotting a calcium trace, stimulus frame rate if plotting running speed
    :param interval_sec: interval in seconds in between labels

    :return: xticks, xticklabels = xticks in units of ophys frames frames, xticklabels in seconds relative
    """
    interval_frames = interval_sec * ophys_frame_rate
    n_frames = len(trace)
    n_sec = n_frames / ophys_frame_rate
    xticks = np.arange(0, n_frames + 1, interval_frames)
    xticklabels = np.arange(0, n_sec + 0.1, interval_sec)
    if not window:
        xticklabels = xticklabels - n_sec / 2
    else:
        xticklabels = xticklabels + window[0]
    if interval_sec >= 1:
        xticklabels = [int(x) for x in xticklabels]
    return xticks, xticklabels


def plot_mean_trace(traces, window=[-4, 8], interval_sec=1, ylabel='dF/F', legend_label=None, color='k', ax=None):
    """
    Function that accepts an array of single trial traces and plots the mean and SEM of the trace, with xticklabels in seconds

    :param traces: array of individual trial traces to average and plot. traces must be of equal length
    :param frame_rate: ophys frame rate if plotting a calcium trace, stimulus frame rate if plotting running speed
    :param y_label: 'dF/F' for calcium trace, 'running speed (cm/s)' for running speed trace
    :param legend_label: string describing trace for legend (ex: 'go', 'catch', image name or other condition identifier)
    :param color: color to plot the trace
    :param interval_sec: interval in seconds for x_axis labels
    :param xlims: range in seconds to plot. Must be <= the length of the traces
    :param ax: if None, create figure and axes to plot. If axis handle is provided, plot is created on that axis

    :return: axis handle
    """
    ophys_frame_rate = 31.  # PROBLEM, shouldn't hard code this here
    if ax is None:
        fig, ax = plt.subplots()
    if len(traces) > 0:
        trace = np.mean(traces, axis=0)
        times = np.arange(0, len(trace), 1)
        sem = (traces.std()) / np.sqrt(float(len(traces)))
        ax.plot(trace, label=legend_label, linewidth=3, color=color)
        ax.fill_between(times, trace + sem, trace - sem, alpha=0.5, color=color)

        xticks, xticklabels = get_xticks_xticklabels(trace, ophys_frame_rate, interval_sec, window=window)
        ax.set_xticks(xticks)
        if interval_sec < 1:
            ax.set_xticklabels(xticklabels)
        else:
            ax.set_xticklabels([int(x) for x in xticklabels])
        ax.set_xlim(0, len(trace))
        ax.set_xlabel('time (sec)')
        ax.set_ylabel(ylabel)
    sns.despine(ax=ax)
    return ax


def plot_flashes_on_trace(ax, window=[-4, 8], go_trials_only=False, omitted=False, flashes=False, alpha=0.25,
                          facecolor='gray'):
    """
    Function to create transparent gray bars spanning the duration of visual stimulus presentations to overlay on existing figure

    :param ax: axis on which to plot stimulus presentation times
    :param window: window of time the trace covers, in seconds
    :param trial_type: 'go' or 'catch'. If 'go', different alpha levels are used for stimulus presentations before and after change time
    :param omitted: boolean, use True if plotting response to omitted flashes
    :param alpha: value between 0-1 to set transparency level of gray bars demarcating stimulus times

    :return: axis handle
    """
    # PROBLEM: shouldn't hard code these things here
    frame_rate = 31.
    stim_duration = .25
    blank_duration = .5
    change_frame = np.abs(window[0]) * frame_rate
    end_frame = (window[1] + np.abs(window[0])) * frame_rate
    interval = blank_duration + stim_duration
    if omitted:
        array = np.arange((change_frame + interval), end_frame, interval * frame_rate)
        array = array[1:]
    else:
        array = np.arange(change_frame, end_frame, interval * frame_rate)
    for i, vals in enumerate(array):
        amin = array[i]
        amax = array[i] + (stim_duration * frame_rate)
        ax.axvspan(amin, amax, facecolor=facecolor, edgecolor='none', alpha=alpha, linewidth=0, zorder=1)
    if go_trials_only:
        alpha = alpha * 3
    else:
        alpha
    array = np.arange(change_frame - ((blank_duration) * frame_rate), 0, -interval * frame_rate)
    for i, vals in enumerate(array):
        amin = array[i]
        amax = array[i] - (stim_duration * frame_rate)
        ax.axvspan(amin, amax, facecolor=facecolor, edgecolor='none', alpha=alpha, linewidth=0, zorder=1)
    return ax


def create_multi_session_mean_df(cache, experiment_ids, conditions=['cell_specimen_id', 'change_image_name'],
                                 flashes=False):
    '''
        Creates a mean response dataframe by combining multiple sessions. 
       
        INPUTS: 
        cache: the cache object for the dataset
        experiment_ids:  a list of experiment_ids for sessions to merge
        conditions: the set of conditions to group by. The first entry should be 'cell_specimen_id'
        flashes: if TRUE, uses the flash_response_df to merge, otherwise uses the trial_response_df

        OUTPUTS
        mega_mdf, a dataframe with index given by the session experiment ids. This allows for easy analysis like:
        mega_mdf.groupby('experiment_id').mean_response.mean()
    '''
    manifest = cache.experiment_table
    mega_mdf = pd.DataFrame()
    # Iterate through experiments
    for experiment_id in experiment_ids:
        # load the session object
        session = cache.get_session(experiment_id)
        print(session.metadata['ophys_experiment_id'])
        # Get the individual session mean_df
        if flashes:
            mdf = get_mean_df(session.flash_response_df, conditions=conditions)
        else:
            mdf = get_mean_df(session.trial_response_df, conditions=conditions)

        # Append metadata
        mdf['experiment_id'] = session.metadata['ophys_experiment_id']
        mdf['experiment_container_id'] = session.metadata['experiment_container_id']
        stage = manifest[manifest.ophys_experiment_id == session.metadata['ophys_experiment_id']].stage_name.values[0]
        mdf['stage_name'] = stage
        mdf['passive'] = parse_stage_for_passive(stage)
        mdf['image_set'] = parse_stage_for_image_set(stage)
        mdf['targeted_structure'] = session.metadata['targeted_structure']
        mdf['imaging_depth'] = session.metadata['imaging_depth']
        mdf['full_genotype'] = session.metadata['full_genotype']
        mdf['cre_line'] = session.metadata['full_genotype'].split('/')[0]
        mdf['retake_number'] = \
            manifest[manifest.ophys_experiment_id == session.metadata['ophys_experiment_id']].retake_number.values[0]

        # Concatenate this session to the other sessions
        mega_mdf = pd.concat([mega_mdf, mdf])

    # Clean up indexes
    mega_mdf = mega_mdf.reset_index()
    mega_mdf = mega_mdf.set_index('experiment_id')
    if 'index' in mega_mdf.keys():
        mega_mdf = mega_mdf.drop(columns=['index'])
    if 'level_0' in mega_mdf.keys():
        mega_mdf = mega_mdf.drop(columns=['level_0'])

    return mega_mdf


def parse_stage_for_passive(stage):
    '''
        Returns TRUE if the stage_name indicates a passive sessions
    '''
    return 'passive' in stage


def parse_stage_for_image_set(stage):
    '''
        Returns the character for the image_set, for example 'A'
    '''
    return stage[15]


def get_active_cell_indices(dff_traces):
    '''
        Returns the ten most active cells.
        Computes active cells by SNR = mean/std over all timepoints. 
    '''
    snr_values = []
    for i, trace in enumerate(dff_traces):
        mean = np.mean(trace, axis=0)
        std = np.std(trace, axis=0)
        snr = mean / std
        snr_values.append(snr)
    active_cell_indices = np.argsort(snr_values)[-10:]
    return active_cell_indices


def compute_lifetime_sparseness(image_responses):
    # image responses should be an array of the trial averaged responses to each image
    # sparseness = 1-(sum of trial averaged responses to images / N)squared / (sum of (squared mean responses / n)) / (1-(1/N))
    # N = number of images
    # after Vinje & Gallant, 2000; Froudarakis et al., 2014
    N = float(len(image_responses))
    ls = ((1 - (1 / N) * ((np.power(image_responses.sum(axis=0), 2)) / (np.power(image_responses, 2).sum(axis=0)))) / (
        1 - (1 / N)))
    return ls
