from typing import List, Tuple
import numpy as np

from scipy import signal


def select_good_channels(lfp: np.ndarray,
                         reference_channels: List[int],
                         noisy_channel_threshold: float
                         ) -> Tuple[np.ndarray, np.ndarray]:
    """Remove reference channels and channels that are too noisy from lfp data.

    Parameters
    ----------
    lfp : numpy.ndarray
        LFP data in the form of: trials x channels x time samples
    reference_channels : List[int]
        Reference channel indices for this probe.
    noisy_channel_threshold : float
        Lowest mean standard deviation that constitutes a "clean" LFP channel

    Returns
    -------
    Tuple[cleaned_lfp, good_indices]
        cleaned_lfp: numpy.ndarray
            LFP where reference and noisy channels have been removed.
            Data still in form of: trials x channel x time samples
        good_indices: numpy.ndarray
            Array of channel indices that are neither reference nor noisy.
    """
    channel_variance = np.mean(np.std(lfp, 2), 0)
    noisy_channels = np.where(channel_variance > noisy_channel_threshold)[0]

    to_remove = np.concatenate((np.array(reference_channels), noisy_channels))
    good_indices = np.delete(np.arange(0, lfp.shape[1]), to_remove)

    # Remove noisy or reference channels (axis=1)
    cleaned_lfp = np.delete(lfp, to_remove, axis=1)

    return (cleaned_lfp, good_indices)


def filter_lfp_channels(lfp: np.ndarray,
                        sampling_rate: float,
                        filter_cuts: List[float],
                        filter_order: int) -> np.ndarray:
    '''Bandpass filter lfp channel data.

    Parameters
    ----------
    lfp : numpy.ndarray
        LFP data to be filtered in the form of:
        trials x channels x time samples
    sampling_rate : float
        Sampling rate for lfp data
    filter_cuts : List[float]
        Low and high cut for bandpass filter
    filter_order : int
        Order for bandpass filter

    Returns
    -------
    filtered_lfp: numpy.ndarray
        LFP that has been bandpassed filtered along the sample axis.
        Still in the form of: trials x channels x time samples
    '''

    wn = (sampling_rate / 2)
    filter_cutoffs = np.array(filter_cuts) / wn
    b, a = signal.butter(filter_order, filter_cutoffs, 'bandpass')
    # Bandpass filter time samples (axis=2)
    filtered_lfp = signal.filtfilt(b, a, lfp, axis=2)

    return filtered_lfp
