import time
t0 = time.time()
import os

from matplotlib.patches import Ellipse
import matplotlib.pyplot as plt
from moviepy.video.io.bindings import mplfig_to_npimage
from moviepy.editor import *
import numpy as np
import collections
import pandas as pd
import sys
import re
import argparse
import logging
from ellipses import LSqEllipse
from google.cloud import storage


ch = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
logger = logging.getLogger('dlc-ellipse-fitting')
logger.setLevel(logging.INFO)
logger.addHandler(ch)
logger.propagate = False

parser = argparse.ArgumentParser()
parser.add_argument("--video_input_file", type=str, required=True, help="path to video file, mp4 or avi")
args = parser.parse_args()

video_file_path = args.video_input_file
h5file_path = video_file_path[:-4] + 'DeepCut_resnet50_universal_eye_trackingJul10shuffle1_1030000.h5'

ellipse_bucket_data_blobname = '{}.h5'.format(os.path.splitext(video_file_path)[0])
ellipse_output_data_file = '/workdir/{}'.format(ellipse_bucket_data_blobname)


client = storage.Client()
src_bucket = client.get_bucket('brain-observatory-dlc-eye-tracking')
tgt_bucket = client.get_bucket('dlc-ellipse-fitting')
blob = src_bucket.get_blob(h5file_path)
blob.download_to_filename(h5file_path)

path_config_file = '/workdir/model/config.yaml'


def fit_ellipse(h5name):
    
    df = pd.read_hdf(h5name).DeepCut_resnet50_universal_eye_trackingJul10shuffle1_1030000

    l_threshold = 0.8 #increased likelihood threshold for points that are allowed in fit
    min_num_points = 6

    # uses https://github.com/bdhammel/least-squares-ellipse-fitting
    # based on the publication Halir, R., Flusser, J.: 'Numerically Stable Direct Least Squares Fitting of Ellipses'

    cr = [] 
    eye = [] 
    pupil = [] 
    
    #new for loop
    loop_t0 = time.time()
    last_loop_time = time.time()
    for j in range(len(df)):
            
        #fit ellipses to the pupil & eye points in 4/25
        
        frac_completed = max(1,float(j))/len(df)
        frac_rem = 1-frac_completed
        tot_time_est = (time.time() - loop_t0)/frac_completed
        progress_str = "{:10.2f} {:10.2f} {:5s} {:10.2f}".format(time.time()-last_loop_time, time.time()-loop_t0, "{0:.0%}".format(frac_completed), tot_time_est)
        logger.info('Ellipse fit: {}'.format(progress_str))
        last_loop_time = time.time()
        
        x_data = df.filter(regex=("cr*")).iloc[j].values[0::3]
        y_data = df.filter(regex=("cr*")).iloc[j].values[1::3]
        l = df.filter(regex=("cr*")).iloc[j].values[2::3]
        try:
            if len(l[l>l_threshold]) >= min_num_points: #at least 6 tracked points for annotation quality data
                lsqe = LSqEllipse() #make fitting object
                lsqe.fit([x_data[l>l_threshold], y_data[l>l_threshold]])
                center, width, height, phi = lsqe.parameters()
                ellipse_dict = {'center_x' : center[0], 'center_y' : center[1], 'width' : width, 'height' : height, 'phi' : phi}
            else:
                ellipse_dict = {'center_x' : np.nan, 'center_y' : np.nan, 'width' : np.nan, 'height' : np.nan, 'phi' : np.nan}
        except Exception as e:
            ellipse_dict = {'center_x' : np.nan, 'center_y' : np.nan, 'width' : np.nan, 'height' : np.nan, 'phi' : np.nan}
            print(e)
        cr.append(ellipse_dict)
        #eye
        x_data = df.filter(regex=("eye*")).iloc[j].values[0::3]
        y_data = df.filter(regex=("eye*")).iloc[j].values[1::3]
        l = df.filter(regex=("eye*")).iloc[j].values[2::3]
        try:
            if len(l[l>l_threshold]) >= min_num_points: #at least 6 tracked points for annotation quality data
                lsqe = LSqEllipse() #make fitting object
                lsqe.fit([x_data[l>l_threshold], y_data[l>l_threshold]])
                center, width, height, phi = lsqe.parameters()
                ellipse_dict = {'center_x' : center[0], 'center_y' : center[1], 'width' : width, 'height' : height, 'phi' : phi}
            else:
                ellipse_dict = {'center_x' : np.nan, 'center_y' : np.nan, 'width' : np.nan, 'height' : np.nan, 'phi' : np.nan}
        except Exception as e:
            ellipse_dict = {'center_x' : np.nan, 'center_y' : np.nan, 'width' : np.nan, 'height' : np.nan, 'phi' : np.nan}
            print(e)
        eye.append(ellipse_dict)  

        
        #pupil
        x_data = df.filter(regex=("pupil*")).iloc[j].values[0::3]
        y_data = df.filter(regex=("pupil*")).iloc[j].values[1::3]
        l = df.filter(regex=("pupil*")).iloc[j].values[2::3]
        try:
            if len(l[l>l_threshold]) >= min_num_points: #at least 6 tracked points for annotation quality data
                lsqe = LSqEllipse() #make fitting object
                lsqe.fit([x_data[l>l_threshold], y_data[l>l_threshold]])
                center, width, height, phi = lsqe.parameters()
                ellipse_dict = {'center_x' : center[0], 'center_y' : center[1], 'width' : width, 'height' : height, 'phi' : phi}
            else:
                ellipse_dict = {'center_x' : np.nan, 'center_y' : np.nan, 'width' : np.nan, 'height' : np.nan, 'phi' : np.nan}
        except Exception as e:
            ellipse_dict = {'center_x' : np.nan, 'center_y' : np.nan, 'width' : np.nan, 'height' : np.nan, 'phi' : np.nan}
            print(e)
        pupil.append(ellipse_dict) 
    
    pd.DataFrame(cr).to_hdf(ellipse_output_data_file, key='cr', mode='w') #overwrite file      
    pd.DataFrame(eye).to_hdf(ellipse_output_data_file, key='eye', mode='a')   
    pd.DataFrame(pupil).to_hdf(ellipse_output_data_file, key='pupil', mode='a')


    blob2 = tgt_bucket.blob(ellipse_bucket_data_blobname)
    blob2.upload_from_filename(filename=ellipse_output_data_file)
    
    
    return cr, eye, pupil
        
initialization_time = time.time() - t0
ellipse_fit_t0 = time.time()
cr, eye, pupil = fit_ellipse(h5file_path)        
ellipse_fit_time = time.time() - ellipse_fit_t0

logger.info('Initialization Time: {}'.format(initialization_time))
logger.info('Ellipse Fit Time: {}'.format(ellipse_fit_time))
logger.info('Total Walltime: {}'.format(time.time()-t0))



